//*****************************************************************************
//
// File Name	: 'Fram.c'
// Title		: Provide FRAM functionality
// Author		: Michael Meyer
// Created		: 15.08.2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: FRAM
//
// 11.059200 MHz external clock
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 

#include "Headers.h"

/** D E C L A R A T I O N S **************************************************/ 

#define FM_WREN				0x06		// Set Write Enable Latch
#define FM_WRDI				0x04		// Write Disable
#define FM_RDSR				0x05		// Read Status Register
#define FM_WRSR				0x01		// Write Status Register
#define FM_READ				0x03		// Read Memory Data
#define FM_WRITE			0x02		// Write Memory Data

// Mode 0, 5,2 MHz SPI CLK
#define	FRAM_CS_ENABLE()	FRAM_PORT &= ~FRAM_CS
#define	FRAM_CS_DISABLE()	FRAM_PORT |= FRAM_CS;

typedef union 
{
	struct {
        uint8_t		b1;
		uint8_t		b2;
		uint8_t		b3;
		uint8_t		b4;
	} Byte;
    uint32_t	 Val;
} Wandler;


/** V A R I A B L E S ********************************************************/ 


/** Global local V A R I A B L E S *******************************************/ 


/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
	
/** C O D E ******************************************************************/ 


//********************************************************/
// FRAMInit
// Initialisation of FRAM functionality
// 
// Parameters:
// none
//********************************************************/
void FRAMInit(void)
{
	// set up ports
	KM_DELAY_MS(12);										// to power up FRAM Chip

	KM_WAIT_SPI();											// wait for SPI bus available
	FRAM_CS_ENABLE();										// activate FRAM SPI interface
	WLANWriteByteSPI(FM_WREN);
	FRAM_CS_DISABLE();										// deactivate FRAM SPI interface

	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_WRSR);
	WLANWriteByteSPI(0);
	FRAM_CS_DISABLE();

	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_RDSR);
	WLANWriteByteSPI(0);									// read
	FRAM_CS_DISABLE();
	KM_RELEASE_SPI();
}


//********************************************************/
// FRAMClear
// Resets the first 50 Variables to 0
// 
// Parameters:
// none
//********************************************************/
void FRAMClear(void)
{
	uint8_t	ii;
	FramMinMax_s	minmax;	
	FramDayVal_s	day;

	for(ii=0;ii<FD_MAX;ii++)
		FRAMWriteLong(0l, ii);
	
	memset(&minmax, 0, sizeof(FramMinMax_s));
	for(ii=0;ii<FD_S_MAX;ii++)
		FRAMWriteBlock((uint8_t *)&minmax, FD_S_START_ADR + (ii * FD_S_ENTRY_SZ), FD_S_ENTRY_SZ);

	memset(&day, 0, sizeof(FramDayVal_s));
	for(ii=0;ii<FD_D_MAX;ii++)
		FRAMWriteBlock((uint8_t *)&minmax, FD_D_START_ADR + (ii * FD_D_ENTRY_SZ), FD_D_ENTRY_SZ);
}

//********************************************************/
// FRAMWriteLong
// Write a number to Fifo
// 
// Parameters:	
//				
//********************************************************/
void FRAMWriteLong(uint32_t Number, uint8_t Pos)
{
	Wandler	Num;

	Num.Val = Number;

	KM_WAIT_SPI();
	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_WREN);
	FRAM_CS_DISABLE();

	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_WRITE);
	WLANWriteByteSPI(0);
	WLANWriteByteSPI(Pos*4);

	WLANWriteByteSPI(Num.Byte.b1);
	WLANWriteByteSPI(Num.Byte.b2);
	WLANWriteByteSPI(Num.Byte.b3);
	WLANWriteByteSPI(Num.Byte.b4);
	
	FRAM_CS_DISABLE();
	KM_RELEASE_SPI();
}

//********************************************************/
// FRAMRead
// Read bytes from Fifo
// 
// Parameters:	
// Return:		
//				
//********************************************************/
uint32_t FRAMReadLong(uint8_t Pos)
{
	Wandler	ret;

	KM_WAIT_SPI();
	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_READ);
	WLANWriteByteSPI(0);
	WLANWriteByteSPI(Pos*4);

	ret.Byte.b1 = WLANWriteByteSPI(0);
	ret.Byte.b2 = WLANWriteByteSPI(0);
	ret.Byte.b3 = WLANWriteByteSPI(0);
	ret.Byte.b4 = WLANWriteByteSPI(0);
	
	FRAM_CS_DISABLE();
	KM_RELEASE_SPI();
	return ret.Val;
}

//********************************************************/
// FRAMWriteBlock
// Write a block to FRAM
// 
// Parameters:	
//				
//********************************************************/
void FRAMWriteBlock(uint8_t *Block, uint16_t StartAdr, uint16_t Size)
{
	KM_WAIT_SPI();
	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_WREN);
	FRAM_CS_DISABLE();

	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_WRITE);
	WLANWriteByteSPI((uint8_t)(StartAdr>>8));
	WLANWriteByteSPI((uint8_t)StartAdr);
	
	while(Size) {
		WLANWriteByteSPI(*Block);
		Block++;
		Size--;
	}
	FRAM_CS_DISABLE();
	KM_RELEASE_SPI();
}

//********************************************************/
// FRAMReadBlock
// Read a block from FRAM
// 
// Parameters:	
//				
//********************************************************/
void FRAMReadBlock(uint8_t *Block, uint16_t StartAdr, uint16_t Size)
{
	KM_WAIT_SPI();
	FRAM_CS_ENABLE();
	WLANWriteByteSPI(FM_READ);
	WLANWriteByteSPI((uint8_t)(StartAdr>>8));
	WLANWriteByteSPI((uint8_t)StartAdr);
	
	while(Size) {
		*Block = WLANWriteByteSPI(0);
		Block++;
		Size--;
	}
	FRAM_CS_DISABLE();
	KM_RELEASE_SPI();
}


