//*****************************************************************************
//
// File Name	: 'KM271Task.c'
// Title		: Main function for handling interface to Logamatic 2105
//				 via selfmade KM271
// Author		: Michael Meyer
// Created		: 13.08.2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: Km
//
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 
#include "Headers.h"


/** D E C L A R A T I O N S **************************************************/ 

#define KM_STACK_SIZE			190							// adjust by checking during runtime
#define	KM_QUEUE_SIZE			2							// 2 items in RX queue, do not change!

#define	KM_STX					0x02						// protocol control bytes
#define	KM_DLE					0x10
#define	KM_ETX					0x03
#define	KM_NAK					0x15

#define	KM_RX_BUF_LEN			10							// max number of RX bytes (8 pure data bytes shoud be sufficient)
#define	KM_TX_BUF_LEN			10							// max number of TX bytes (8 pure data bytes shoud be sufficient)

#define	KM_RX_RESYNC			0							// unknown state re-sync by wait for STX
#define	KM_RX_IDLE				1							// states for RX interrupt routine
#define	KM_RX_ON				2							// Block reception started
#define	KM_RX_DLE				3							// DLE doubling
#define	KM_RX_ETX				4							// end detected
#define	KM_RX_BCC				5							// verify block

#define	KM_TX_IN_BLOCK			0							// currently sending byte out of block
#define	KM_TX_DLE_DOUB			1							// DLE doubling in progress

#define	KM_TX_DLE				0							// DLE to be sent
#define	KM_TX_ETX				1							// ETX to be sent
#define	KM_TX_BCC				2							// BCC to be sent
#define	KM_TX_DONE				3							// Done

#define	KM_NUM_EVENTS			(sizeof(KmParamR)/sizeof(KmParamR_s)) // Number of known heating events/parmeters

// KMTaskStatus
#define	KM_TSK_START			0							// switch to logging mode
#define	KM_TSK_LG_CMD			1							// receive confirmation from KM
#define	KM_TSK_LOGGING			2							// Logging active

#define	KM_IN_FLASH				1							// Block is stored in PROGMEM
#define	KM_IN_RAM				0							// Block is stored in RAM

typedef struct {											// Rx structure for one rx block
	uint8_t	volatile			len;						// length of data in buffer
	uint8_t	volatile 			buf[KM_RX_BUF_LEN];			// received bytes without "10 03 bcc"
} KmRx_s;

typedef struct {											// Tx structure for one Tx block
	uint8_t	volatile 			len;						// length of data to be transmitted
	uint8_t						buf[KM_TX_BUF_LEN];			// block to be transmitted
} KmTx_s;

typedef struct {
	uint8_t	volatile			st_block:4;					// status during block transmission
	uint8_t	volatile			st_end:4;					// status during end
} KmTxStat_s;

/** V A R I A B L E S ********************************************************/ 
xSemaphoreHandle 				qSPIMutex;					// the SPI Mutex controls acces to SPI bus
uint8_t							KmKesselTemp;


// known commands to KM271, to be used with KmStartTx()
const uint8_t KmCSTX[]  	PROGMEM = {KM_STX};				// STX Command
const uint8_t KmCDLE[]  	PROGMEM = {KM_DLE};				// DLE Response
const uint8_t KmCLogMode[]  PROGMEM = {0xEE, 0x00, 0x00};	// Switch to Log Mode


// creating some tables to hold retrieved heating parameters / events
// strings are hold in PROGMEM, the actaul data in RAM 
// content of different tables is linked by the index

// Parameter texts stored in PROGMEM
const char KmP0[]   PROGMEM = "HK Betriebswerte 1";			// 0x8000
const char KmAtP0[] PROGMEM = "HBW1";						

const char KmP1[]   PROGMEM = "HK Betriebswerte 2";			// 0x8001
const char KmAtP1[] PROGMEM = "HBW2";						

const char KmP2[]   PROGMEM = "Vorlaufsolltemp";			// 0x8002
const char KmAtP2[] PROGMEM = "HVLS";						

const char KmP3[]   PROGMEM = "Vorlaufisttemp";				// 0x8003
const char KmAtP3[] PROGMEM = "HVLI";						

const char KmP4[]   PROGMEM = "Raumsolltemp";				// 0x8004
const char KmAtP4[] PROGMEM = "HRS";						

const char KmP5[]   PROGMEM = "Raumisttemp";				// 0x8005
const char KmAtP5[] PROGMEM = "HRI";						

const char KmP6[]   PROGMEM = "EinschaltOptZeit";			// 0x8006
const char KmAtP6[] PROGMEM = "HEOZ";						

const char KmP7[]   PROGMEM = "AusschaltOptZeit";			// 0x8007
const char KmAtP7[] PROGMEM = "HAOZ";						

const char KmP8[]   PROGMEM = "Pumpe";						// 0x8008
const char KmAtP8[] PROGMEM = "HPUMP";						

const char KmP9[]   PROGMEM = "Mischerstellung";			// 0x8009
const char KmAtP9[] PROGMEM = "HMSCH";						

const char KmP10[]   PROGMEM = "Heizkennlinie +10";			// 0x800c
const char KmAtP10[] PROGMEM = "HK10";						

const char KmP11[]   PROGMEM = "Heizkennlinie 0";			// 0x800d
const char KmAtP11[] PROGMEM = "HK0";						

const char KmP12[]   PROGMEM = "Heizkennlinie -10";			// 0x800e
const char KmAtP12[] PROGMEM = "HKM10";						

const char KmP13[]   PROGMEM = "WW Betriebswerte 1";		// 0x8424
const char KmAtP13[] PROGMEM = "HWWBW1";						

const char KmP14[]   PROGMEM = "WW Betriebswerte 2";		// 0x8425
const char KmAtP14[] PROGMEM = "HWWBW2";						

const char KmP15[]   PROGMEM = "WW Solltemp";				// 0x8426
const char KmAtP15[] PROGMEM = "HWWST";						

const char KmP16[]   PROGMEM = "WW Isttemp";				// 0x8427
const char KmAtP16[] PROGMEM = "HWWIT";						

const char KmP17[]   PROGMEM = "WW Optzeit";				// 0x8428
const char KmAtP17[] PROGMEM = "HWWOZ";						

const char KmP18[]   PROGMEM = "WW Pumpen";					// 0x8429
const char KmAtP18[] PROGMEM = "HWWLP";						

const char KmP19[]   PROGMEM = "Kesselvorlaufsolltemp";		// 0x882a
const char KmAtP19[] PROGMEM = "HKVST";						
   	
const char KmP20[]   PROGMEM = "Kesselvorlaufisttemp";		// 0x882b
const char KmAtP20[] PROGMEM = "HKVIT";						

const char KmP21[]   PROGMEM = "Brennereinschalttemp";		// 0x882c
const char KmAtP21[] PROGMEM = "HBEST";						

const char KmP22[]   PROGMEM = "Brennerausschalttemp";		// 0x882d
const char KmAtP22[] PROGMEM = "HBAST";						

const char KmP23[]   PROGMEM = "Kesselintegral 1";			// 0x882e
const char KmAtP23[] PROGMEM = "HKI1";						

const char KmP24[]   PROGMEM = "Kesselintegral 2";			// 0x882f
const char KmAtP24[] PROGMEM = "HKI2";						

const char KmP25[]   PROGMEM = "Kesselfehler";				// 0x8830
const char KmAtP25[] PROGMEM = "HKFEH";						

const char KmP26[]   PROGMEM = "Kesselbetrieb";				// 0x8831
const char KmAtP26[] PROGMEM = "HKBET";						

const char KmP27[]   PROGMEM = "Brenneransteuerung";		// 0x8832
const char KmAtP27[] PROGMEM = "HBREN";						

const char KmP28[]   PROGMEM = "Abgastemperatur";			// 0x8833
const char KmAtP28[] PROGMEM = "HABGS";						

const char KmP29[]   PROGMEM = "Brennerlaufz. Std 2";		// 0x8836
const char KmAtP29[] PROGMEM = "HBLZ2";						

const char KmP30[]   PROGMEM = "Brennerlaufz. Std 1";		// 0x8837
const char KmAtP30[] PROGMEM = "HBLZ1";						

const char KmP31[]   PROGMEM = "Brennerlaufz. Std 0";		// 0x8838
const char KmAtP31[] PROGMEM = "HBLZ0";						

const char KmP32[]   PROGMEM = "Aussentemp";				// 0x893c
const char KmAtP32[] PROGMEM = "HAUSS";						

const char KmP33[]   PROGMEM = "Aussentemp gedaempft";		// 0x893d
const char KmAtP33[] PROGMEM = "HAUSG";						

const char KmP34[]   PROGMEM = "Hauptversion";				// 0x893e
const char KmAtP34[] PROGMEM = "HHVER";						

const char KmP35[]   PROGMEM = "Unterversion";				// 0x893f
const char KmAtP35[] PROGMEM = "HUVER";						

const char KmP36[]   PROGMEM = "Modul";						// 0x8940
const char KmAtP36[] PROGMEM = "HMOD";						

   	   	   	   	
// fix part of the parmeter list, located in ROM
const KmParamR_s KmParamR[] PROGMEM = {
	{0x00, KmP0, KmAtP0, KM_PAR_BITFIELD },					// 0x8000
	{0x01, KmP1, KmAtP1, KM_PAR_BITFIELD },					// 0x8001
	{0x02, KmP2, KmAtP2, KM_PAR_TEMP },						// 0x8002
	{0x03, KmP3, KmAtP3, KM_PAR_TEMP },						// 0x8003
	{0x04, KmP4, KmAtP4, KM_PAR_TEMP_HALF },				// 0x8004
	{0x05, KmP5, KmAtP5, KM_PAR_TEMP_HALF },				// 0x8005
	{0x06, KmP6, KmAtP6, KM_PAR_MINUTES },					// 0x8006
	{0x07, KmP7, KmAtP7, KM_PAR_MINUTES },					// 0x8007
	{0x08, KmP8, KmAtP8, KM_PAR_PERCENT },					// 0x8008
	{0x09, KmP9, KmAtP9, KM_PAR_PERCENT },					// 0x8009
	{0x0c, KmP10, KmAtP10, KM_PAR_TEMP },					// 0x800c
	{0x0d, KmP11, KmAtP11, KM_PAR_TEMP },					// 0x800d
	{0x0e, KmP12, KmAtP12, KM_PAR_TEMP },					// 0x800e
	{0x24, KmP13, KmAtP13, KM_PAR_BITFIELD },				// 0x8424
	{0x25, KmP14, KmAtP14, KM_PAR_BITFIELD },				// 0x8425
	{0x26, KmP15, KmAtP15, KM_PAR_TEMP },					// 0x8426
	{0x27, KmP16, KmAtP16, KM_PAR_TEMP },					// 0x8427
	{0x28, KmP17, KmAtP17, KM_PAR_MINUTES },				// 0x8428
	{0x29, KmP18, KmAtP18, KM_PAR_BITFIELD },				// 0x8429
	{0x2a, KmP19, KmAtP19, KM_PAR_TEMP },					// 0x882a
	{0x2b, KmP20, KmAtP20, KM_PAR_TEMP },					// 0x882b
	{0x2c, KmP21, KmAtP21, KM_PAR_TEMP },					// 0x882c
	{0x2d, KmP22, KmAtP22, KM_PAR_TEMP },					// 0x882d
	{0x2e, KmP23, KmAtP23, KM_PAR_NUMBER },					// 0x882e
	{0x2f, KmP24, KmAtP24, KM_PAR_NUMBER },					// 0x882f
	{0x30, KmP25, KmAtP25, KM_PAR_BITFIELD },				// 0x8830
	{0x31, KmP26, KmAtP26, KM_PAR_BITFIELD },				// 0x8831
	{0x32, KmP27, KmAtP27, KM_PAR_BITFIELD },				// 0x8832
	{0x33, KmP28, KmAtP28, KM_PAR_TEMP },					// 0x8833
	{0x36, KmP29, KmAtP29, KM_PAR_NUMBER },					// 0x8836
	{0x37, KmP30, KmAtP30, KM_PAR_NUMBER },					// 0x8837
	{0x38, KmP31, KmAtP31, KM_PAR_NUMBER },					// 0x8838
	{0x3c, KmP32, KmAtP32, KM_PAR_TEMP_NEG },				// 0x893c
	{0x3d, KmP33, KmAtP33, KM_PAR_TEMP_NEG },				// 0x893d
	{0x3e, KmP34, KmAtP34, KM_PAR_NUMBER },					// 0x893e
	{0x3f, KmP35, KmAtP35, KM_PAR_NUMBER },					// 0x893f
	{0x40, KmP36, KmAtP36, KM_PAR_NUMBER },					// 0x8940
};

// actual parameter values, located in RAM; index linked to KmParamR
uint8_t	KmParamD[KM_NUM_EVENTS];

/** Global local V A R I A B L E S *******************************************/ 
xTaskHandle 					xKmTaskHandle;				// handle for the KM task
xQueueHandle 					qKmQueue;					// the handle to the KM queue
uint8_t volatile				KmRxStatus;					// status in Rx interrupt
uint8_t volatile				KmRxBcc;					// BCC value for Rx Block
uint8_t volatile				KmTxBcc;					// BCC value for Tx Block
uint8_t volatile				KmTxIndex;					// Index value for Tx Block
KmTxStat_s						KmTxStatus;					// status in Tx interrupt
uint8_t volatile				KmRxBufIndex;				// Index of current RxBuf (0 or one)
KmRx_s 							KmRxBuf[KM_QUEUE_SIZE];		// Rx block storage, two blocks as there are cases with 2 consecutive answers from KM
KmTx_s 							TxBuf;						// Tx storage in task
#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
uint16_t volatile				KmRxErr;					// count number of RX Errors
#endif

/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
portTASK_FUNCTION_PROTO(KmTask, pvParameters);				// the task handling the 3964 protocol (partly)
void KmUartInit(void);										// initialises the UART (2400 8N1) and enables RX
void KmStartTx(const uint8_t *Block, uint8_t len, uint8_t InFlash);		// initiates sending of block
void ParseBlock(KmRx_s *RxBuf);								// analyses one received block


/** C O D E ******************************************************************/ 


//********************************************************/
// UART RX Interrupt routine
// This interrupt is activated 
// when there is a character available in RX
// 
// STX, DLE, NAK are given directly to receiving task
// Blocks are provided to receiving task as pure data, without DLE,ETX,BCC
// On BCC errors, requesting re-transmission of block is handled here
//
// In case of any error, the routine is relying on the KM271 
// to recover again, just by waiting for the next STX from KM271.
// So no timeout or extended error recovery is implemented.
// But this seems to be fine...
// 
// Parameters:
// none
//********************************************************/
ISR (USART0_RX_vect)  
{
	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE; 
	uint8_t		rxbyte;										// to store rx byte in this interrupt
	KmRx_s		*bufPtr;									// pointer to current RxBuf structure

	if(UCSR0A & ((1<<FE0) | (1<<DOR0)))	{					// any problems during reception?
		rxbyte = UDR0;										// get the received byte, clear errors
		KmRxStatus = KM_RX_RESYNC;							// try to resynchronise
		return;												// ignore the byte, timeout and STX will handle it
	}

	rxbyte = UDR0;											// get the received byte

	bufPtr = &KmRxBuf[KmRxBufIndex];						// set pointer to current rx Buffer
	
	LED2_TOGGLE;											// indicate some activity
	KmRxBcc ^= rxbyte;										// calculate BCC
	switch(KmRxStatus) {
		case KM_RX_RESYNC:									// unknown state, discard everthing but STX
			if(rxbyte == KM_STX) {							// react on STX only to re-synchronise
				bufPtr->buf[0] = KM_STX;					// store current STX
				bufPtr->len = 1;							// set length
				KmRxStatus = KM_RX_IDLE;					// sync done, now continue to receive
				xQueueSendFromISR(qKmQueue, &bufPtr, &xHigherPriorityTaskWoken);	// send it to task
				KmRxBufIndex^=1;							// toggle rx buffer index, switch to next rx buffer
			}
			break;
		case KM_RX_IDLE:									// start of block or command
			bufPtr->buf[0] = rxbyte;						// store current byte
			bufPtr->len = 1;								// initialise length
			KmRxBcc = rxbyte;								// reset BCC
			if((rxbyte == KM_STX) || (rxbyte == KM_DLE) || (rxbyte == KM_NAK)) {	// give STX, DLE, NAK directly to caller
				xQueueSendFromISR(qKmQueue, &bufPtr, &xHigherPriorityTaskWoken);	// send it to task
				KmRxBufIndex^=1;							// toggle rx buffer index, switch to next rx buffer
			} else {										// whole block will follow
				KmRxStatus = KM_RX_ON;						// more data to follow, start collecting
			}
			break;											
		case KM_RX_ON:										// block reception ongoing
			if(rxbyte == KM_DLE) {							// handle DLE doubling
				KmRxStatus = KM_RX_DLE;						// discard first received DLE, could be doubling or end of block, check in next state
				break;										// quit here without storing
			}
			if(bufPtr->len >= KM_RX_BUF_LEN) {				// check allowed block len, if too long, re-sync
				#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
					KmRxErr++;
				#endif
				KmRxStatus = KM_RX_RESYNC;					// enter re-sync
				break;										// do not save data beyond array border
			}
			bufPtr->buf[bufPtr->len] = rxbyte;				// no DLE -> store regular, current byte
			bufPtr->len++;									// adjust length in rx buffer
			break;
		case KM_RX_DLE:										// entered when one DLE was already received
			if(rxbyte == KM_DLE) {							// double DLE?
				if(bufPtr->len >= KM_RX_BUF_LEN) {			// check allowed block len, if too long, re-sync
					#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
						KmRxErr++;
					#endif
					KmRxStatus = KM_RX_RESYNC;				// enter re-sync
					break;									// do not save data beyond array border
				}
				bufPtr->buf[bufPtr->len] = rxbyte;			// yes -> store this DLE as valid part of data
				bufPtr->len++;								// adjust length in rx buffer
				KmRxStatus = KM_RX_ON;						// continue to receive block
			} else {										// this should be ETX now
				if(rxbyte == KM_ETX) {						// really? then we are done, just waiting for BCC
					KmRxStatus = KM_RX_BCC;					// receive BCC and verify it
				} else {
					#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
						KmRxErr++;
					#endif
					KmRxStatus = KM_RX_RESYNC;				// something wrong, just try to restart 
				}
			}
			break;
		case KM_RX_BCC:										// last stage, BCC verification, "received BCC" ^ "calculated BCC" shall be 0 
			if(!KmRxBcc) {									// block is valid
				xQueueSendFromISR(qKmQueue, &bufPtr, &xHigherPriorityTaskWoken);	// give block to caller 
				KmRxBufIndex^=1;							// toggle rx buffer index, switch to next rx buffer
			} else {
				#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
					KmRxErr++;
				#endif
				UDR0 = KM_NAK;								// send NAK, ask for re-sending the block
			}
			KmRxStatus = KM_RX_IDLE;						// wait for next data or re-sent block
			break;
	}

	// wakeup waiting task, if higher prio
	if(xHigherPriorityTaskWoken == pdTRUE) {
		taskYIELD();
    }
}

//********************************************************/
// UART TX Interrupt routine
// This interrupt is activated, 
// when there is no charcater in TX buf of UART
// 
// The interrupt handles doubling of 0x10 in data and
// closing the block with DLE, ETX, BCC.
//
// Blocks with a single byte are assumed to be control blocks
// with STX, DLE, NAK only. These blocks will not be 
// closed with DLE, ETX, BCC.
// 
// Parameters:
// none
//********************************************************/
ISR (USART0_UDRE_vect)  
{
	if(TxBuf.len > 1) {										// still some data to send
		switch(KmTxStatus.st_block) {
			case KM_TX_IN_BLOCK:
				KmTxIndex++;								// switch to next byte
				KmTxBcc ^= TxBuf.buf[KmTxIndex];			// adjust BCC
				if(TxBuf.buf[KmTxIndex] != KM_DLE) {		// if != 0x10 byte -> indicate byte is send
					TxBuf.len--;							// done with regular byte
				} else {									// if 0x10 byte
					KmTxStatus.st_block = KM_TX_DLE_DOUB;	// do not decrease len, but send another 0x10
				}
				UDR0 = TxBuf.buf[KmTxIndex];				// send next byte	
				break;
			case KM_TX_DLE_DOUB:
				KmTxBcc ^= KM_DLE;							// adjust BCC, even for double 0x10
				KmTxStatus.st_block = KM_TX_IN_BLOCK;		// back to regular block operation
				TxBuf.len--;								// now indicate both 0x10 are send
				UDR0 = KM_DLE;								// send the additional 0x10
				break;
		}
	} else {												// done, no more data byte to send
		if(KmTxIndex == 0) {								// len <= 1 and index == 0 means 1 byte buffer (STX, DLE, NAK)
			UCSR0B &= ~(1<<UDRIE0);							// no more character to sent: stop interrupts
		} else {											// we have reached end of regular block
			switch(KmTxStatus.st_end) {						// now step through sending of end of block
				case KM_TX_DLE:								// first step: send DLE
					KmTxStatus.st_end = KM_TX_ETX;			// switch to next state
					KmTxBcc ^= KM_DLE;						// adjust BCC
					UDR0 = KM_DLE;							// send DLE
					break;
				case KM_TX_ETX:
					KmTxStatus.st_end = KM_TX_BCC;			// switch to next state
					KmTxBcc ^= KM_ETX;						// adjust BCC
					UDR0 = KM_ETX;							// send ETX
					break;
				case KM_TX_BCC:
					KmTxStatus.st_end = KM_TX_DONE;			// switch to next state
					UDR0 = KmTxBcc;							// send BCC
					break;
				case KM_TX_DONE:
					UCSR0B &= ~(1<<UDRIE0);					// no more character: stop interrupts
					break;
			}
		}
	}
}

//********************************************************/
// Km271TaskInit
// Initialisation of KM Task, called once from main
// 
// Parameters:
// none
//********************************************************/
uint8_t Km271TaskInit(void)
{
	portENTER_CRITICAL();
	{
		qKmQueue = xQueueCreate(KM_QUEUE_SIZE, sizeof(KmRx_s *));	// create Km queue to receive blocks from RX Int
		xTaskCreate(KmTask, (signed portCHAR *)"K", KM_STACK_SIZE, NULL, KM_ISV_PRIO, &xKmTaskHandle);	// create IntServ task	
		qSPIMutex = xSemaphoreCreateMutex();						// create Mutex in order to synchronise access to SPI interface
	}
	portEXIT_CRITICAL();

	return TRUE; 
}


//********************************************************/
// Km271Task
// Handles communication with KM271 and Logamatic 2105
// 
// Parameters:
// none
//********************************************************/
portTASK_FUNCTION(KmTask, pvParameters)
{
	(void)		pvParameters;								// Just to stop compiler warnings.
	uint8_t		taskStatus;									// status of task
	KmRx_s		*rxBuf;										// pointer to the rx data

	// immediately suspend timer task in order to ensure proper initialisation
	// will be resumed after initialisation
	vTaskSuspend(xTmrTaskHandle);
	vTaskSuspend(xUpdTaskHandle);

	WLANSpiInit();											// initialise SPI interface
	FRAMInit();												// initialise FRAM
	WLANInit();												// initialise WLAN

	// FRAMClear();											// just use once to initialise variables in FRAM!

	// Everything up and runing. Now we can resume the timer task and the update task.
	vTaskResume(xUpdTaskHandle);
	vTaskResume(xTmrTaskHandle);

	// give KM271 some time to boot up, before we enable reception
	// give WLAN some time to retrieve time/date
	vTaskDelay(20000 / portTICK_RATE_MS);

	// communication path should be clean until now
	KmUartInit();											// now we are ready to receive RX interrupts

	taskStatus = KM_TSK_START;								// prepare state interpreter in task
	#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
		KmRxErr = 0;
	#endif

	while(1) {
		xQueueReceive(qKmQueue, &rxBuf, portMAX_DELAY);		// wait for RX data
		switch(taskStatus) {
			case KM_TSK_START:								// we need to switch to logging mode, first
				switch(rxBuf->buf[0]) {
					case KM_STX:							// first step: wait for STX
						KmStartTx(KmCSTX, sizeof(KmCSTX), KM_IN_FLASH);	// send STX to KM271
						break;
					case KM_DLE:							// DLE received, KM ready to receive command
						KmStartTx(KmCLogMode, sizeof(KmCLogMode), KM_IN_FLASH);	// send logging command
						taskStatus = KM_TSK_LG_CMD;			// switch to check for logging mode state
						break;
				}
				break;
			case KM_TSK_LG_CMD:								// check if logging mode is accepted by KM
				if(rxBuf->buf[0] != KM_DLE) {				// no DLE, not accepted, try again
					taskStatus = KM_TSK_START;				// back to START state
				} else {
					taskStatus = KM_TSK_LOGGING;			// command accepted, ready to log!
				}
				break;
			case KM_TSK_LOGGING:							// we have reached logging state
				if(rxBuf->buf[0] != KM_STX) {				// if not STX, it should be valid data
					ParseBlock(rxBuf);						// handle data block
				}
				KmStartTx(KmCDLE, sizeof(KmCDLE), KM_IN_FLASH);	// confirm handling of block by sending DLE
				break;
		}
	}
} 


//********************************************************/
// KmUartInit
// Initialises the UART for KM271 communication
// 2400,8N1, no HW Flow Control
// 
// Parameters:
// none
//********************************************************/
void KmUartInit(void)
{
	KmRxBufIndex = 0;										// initialise RxBufIndex
	UBRR0 = (uint16_t)(287);								// Set baud rate, 2400 at 11.059200 MHz 
	UCSR0B = 0x00;											// initialise UCSR0B
	UCSR0C = 0x06;											// Set frame format: 8N1
	UCSR0B=(1<<RXEN0)|(1<<TXEN0)|(1<<RXCIE0);				// Enable transmitter, receiver and receiver Interrupt
}

//********************************************************/
// KmStartTx
// Start sending a block of data
// Adding DLE,ETX,BCC and DLE doubling is handled by TX Int
// 
// Parameters:
// Pointer to location with command
// If InFlash != 0, location is assumed to be in Flash 
//********************************************************/
void KmStartTx(const uint8_t *Block, uint8_t len, uint8_t InFlash)	// initiates sending of block
{
	if(InFlash) {
		memcpy_P(TxBuf.buf, Block, len);  						// get command out of flash 
	} else {
		memcpy(TxBuf.buf, Block, len);  						// copy block from RAM
	}
	TxBuf.len = len;										// set len
	KmTxIndex = 0;											// reset index for TX interrupt routine
	KmTxStatus.st_block = KM_TX_IN_BLOCK;					// reset state machine for TX interrupt
	KmTxStatus.st_end = KM_TX_DLE;							// reset state machine for TX interrupt
	UDR0 = KmTxBcc = TxBuf.buf[0];							// write first byte to UART (before enabling INT!) 
	UCSR0B |= (1<<UDRIE0);									// Enable Tx Interrupt, rest is done by TX int
}

//********************************************************/
// ParseBlock
// Parse a block for events and provide index of event to
// Update task.
// Ignore all unknown events
// 
// Parameters:
// none
//********************************************************/
void ParseBlock(KmRx_s *RxBuf)
{
	uint8_t			index;
	
	if(RxBuf->buf[0] & 0x80) {								// highest bit indicates event
		for(index = 0; index < KM_NUM_EVENTS; index++) {	// go through event table
			if(RxBuf->buf[1] == pgm_read_byte(&KmParamR[index].Id)) {	// look for known event
				KmParamD[index] = RxBuf->buf[2];			// found, store value
				xQueueSend(qUpdQueue, &index, 0);			// send index to Update Task
				break;										// there is only one event per block -> quit here
			}
		}
	}
}

