//*****************************************************************************
//
// File Name	: 'main.c'
// Title		: Main function for handling SDCard, MP3 and WLAN
// Author		: Michael Meyer
// Created		: August 2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: Man
//
// 11.059200 MHz external clock
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 
#include "headers.h"

/** D E C L A R A T I O N S **************************************************/ 

/** V A R I A B L E S ********************************************************/ 
#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
uint8_t	ManResCnt EEMEM = 0;								// number of resets
#endif

/** Global local V A R I A B L E S *******************************************/ 

/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
void	ManInitIO(void);


/** C O D E ******************************************************************/ 

//********************************************************/
// main
// Initialisation of hardware and mainloop
// 
// Parameters:
// none
//********************************************************/
int main(void)
{
 	ManInitIO();											// reset all ports to defined values 

#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
	portDISABLE_INTERRUPTS();								// update reset counter
	eeprom_write_byte(&ManResCnt, eeprom_read_byte(&ManResCnt)+1);
	portENABLE_INTERRUPTS();
#endif

	// Start all tasks
	TmrTaskInit();											// initialise timer task

	UpdTaskInit();											// initialise update task

	Km271TaskInit();										// initialise KM271 interface task

	// start the scheduler
	vTaskStartScheduler();									// will never return
	return 0;												// should never come here
}


/********************************************************/
// ManInitIO
// Initialises the hardware
// 
// Parameters:
// none
//********************************************************/
void ManInitIO(void)
{
	// set up ports
	DDRA = PORTA_DIR;										// initialise PORTA
	PORTA = PORTA_INIT;
	
	DDRB = PORTB_DIR;										// initialise PORTB
	PORTB = PORTB_INIT;

	DDRC = PORTC_DIR;										// initialise PORTC
	PORTC = PORTC_INIT;
	
	DDRD = PORTD_DIR;										// initialise PORTD
	PORTD = PORTD_INIT;

	ACSR |= (1<<ACD);										// disable Analog Comperator

	LED1_OFF;												// switch all LEDs off
	LED2_OFF;
	LED3_OFF;

	// now start to initialise the special functionalities

	// FRAM SPI init
	FRAM_PORT |= FRAM_CS;									// put CS to 1
	FRAM_DIR |= FRAM_CS;									// define CS as output

	// SDCARD SPI init
	SDCARD_PORT	|= SDCARD_CS;								// put CS to 1
	SDCARD_DIR |= SDCARD_CS;								// define CS as output

	// WLAN HW interface setup
	WLAN_SPIR_PORT |= WLAN_SPI_RDY;							// enable pull-up for SPI READY
	WLAN_SPIR_DIR &=  ~WLAN_SPI_RDY;						// define DATA_READY as input

	WLANCS_PORT |=	WLAN_CS;								// put WLAN_CS to 1
	WLANCS_DIR |=  WLAN_CS;									// define CS as output

	WLANRES_PORT |=	WLAN_RES;								// put WLAN RESET to 1
	WLANRES_DIR |=  WLAN_RES;								// define RESET as output

	// all other WLAN inputs with pullup
	WLANICHIP_PORT |= WLAN_ICHIP_RDY;						// enable pull up for WLAN iChip Ready
	WLANICHIP_DIR &= ~WLAN_ICHIP_RDY;						// input

	WLANSTAT_PORT |= WLAN_DATA_RDY;							// enable pull up for WLAN Data Ready
	WLANSTAT_DIR &= ~(WLAN_ICHIP_RDY | WLAN_DATA_RDY);		// input

	WLAN_DIS_DATA_INT;										// disable interrupt
	WLAN_DATA_HEDGE;										// configure rising edge data available interrupt
	WLAN_DATA_CLR_INT;										// clear interrupt flag, as interrupt can be triggered by setting edge

	WLAN_DIS_SPIR_INT;										// disable interrupt
	WLAN_SPIR_HEDGE;										// configure rising edge WLAN SPI interrupt
	WLAN_SPIR_CLR_INT;										// clear interrupt flag, as interrupt can be triggered by setting edge

	configure_pin_available();								// by using SD card function (slow interface first)
	configure_pin_locked();									// full speed SPI interface will be enabled by WLANInit()

	configure_pin_mosi();									// set pins for SD Card interface
	configure_pin_sck();
	configure_pin_ss();
	configure_pin_miso();

	unselect_card();										// set SD Cras CS to high

	sei();													// all maskable enable interrupts
}

//********************************************************/
// Stack Checking helper
// This is called by RTOS when Stackoverflow occured
// 
// 
// Parameters:
// none
//********************************************************/
#if( configCHECK_FOR_STACK_OVERFLOW > 0 )
void vApplicationStackOverflowHook( xTaskHandle *pxTask, signed portCHAR *pcTaskName) 
{
	cli();
	while(1);
}
#endif

