//*****************************************************************************
//
// File Name	: 'TmrTask.c'
// Title		: Main function for Timer actions
// Author		: Michael Meyer
// Created		: 15.08.2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: Tmr
//
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 
#include "Headers.h"


/** D E C L A R A T I O N S **************************************************/ 

#define TMR_STACK_SIZE			190							// 420
#define TMR_CYCLE_RATE_MS       (1000 / portTICK_RATE_MS)	// one second cycle time


// some flags
typedef struct {
	uint8_t	Min1:1;											// 1 minute timer in progress
	uint8_t	Min10:1;										// 10 minute timer in progress
	uint8_t	Sync:1;											// sync of time in progress
	uint8_t	NewDay:1;										// new day indication in progress
} TFlags_s;

/** V A R I A B L E S ********************************************************/ 
xTaskHandle 					xTmrTaskHandle;				// this task
TmrTime_s volatile				TmrTime;					// the actual time

/** Global local V A R I A B L E S *******************************************/ 
static const uint8_t			MDays[] = { 29, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
//some flags...
TFlags_s						TmrFlags;

/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
portTASK_FUNCTION_PROTO(TimerTask, pvParameters);	
void TmrIncClock(volatile TmrTime_s *TPtr);


/** C O D E ******************************************************************/ 


//********************************************************/
// TmrTaskInit
// Initialisation of Timer Task, called once from main
// 
// Parameters:
// none
//********************************************************/
uint8_t TmrTaskInit(void)
{
	portENTER_CRITICAL();
	{
		// create Timer task	
		xTaskCreate(TimerTask, (signed portCHAR *)"T", TMR_STACK_SIZE, NULL, TMR_TMR_PRIO, &xTmrTaskHandle);
	}
	portEXIT_CRITICAL();

	return TRUE; 
}

//********************************************************/
// TimerTask
// Handles Timer activities
// 
// Parameters:
// none
//********************************************************/
portTASK_FUNCTION(TimerTask, pvParameters)
{
 	(void)			pvParameters;							// Just to stop compiler warnings.
	portTickType 	xLastWakeTime;	
	uint8_t			sender;
	uint8_t			ii;

	ii = 0;
	TmrFlags.Min1 = 0;
	TmrFlags.Min10 = 0;
	TmrFlags.Sync = 0;
	TmrFlags.NewDay = 0;

	// Initialise the xLastWakeTime variable with the current time.
	xLastWakeTime = xTaskGetTickCount();

	while(1) {
		vTaskDelayUntil(&xLastWakeTime,TMR_CYCLE_RATE_MS);	// wake up every second
		
		LED1_TOGGLE;										// show some sign of life to outside world

		portENTER_CRITICAL();
		{
			TmrIncClock(&TmrTime);							// increment system time 
		}
		portEXIT_CRITICAL();

		// every midnight, keep this as first entry, order is important for saving events in Update
		if((TmrTime.hour == 0) && (TmrTime.minute == 0) && !TmrFlags.NewDay) { // handle today/yesterday
			TmrFlags.NewDay = 1;							// indicate ongoing New Day alert
			sender = UPD_S_DAY_TIMER;						// its an New Day alert
			xQueueSend(qUpdQueue, &sender, 0);				// put in queue
		} else {
			if(TmrTime.minute != 0) {
				TmrFlags.NewDay = 0;						// reset ongoing sync when minute is over
			}
		}

		// every night at 03:05 
		if((TmrTime.hour == 3) && (TmrTime.minute == 5) && !TmrFlags.Sync) {	// sync time every day at 03:05 Uhr
			TmrFlags.Sync = 1;								// indicate ongoing sync
			sender = UPD_S_SYNC_TIMER;						// its  sync timer
			xQueueSend(qUpdQueue, &sender, 0);				// put in queue
		} else {
			if(TmrTime.minute != 5) {						
				TmrFlags.Sync = 0;							// reset ongoing sync when minute is over
			}
		}

		// every 10 seconds
		if(++ii>=10) {										// 10 second tick for Update Task
			sender = UPD_S_10STIMER;						// its 10 secound timer
			xQueueSend(qUpdQueue, &sender, 0);				// wake up Update Task 	
			ii = 0;
		}

		// every 1 minutes 
		if(!TmrTime.second  && !TmrFlags.Min1) {			// every 1 minute
			TmrFlags.Min1 = 1;								// indicated saved
			sender = UPD_S_1MTIMER;	
			xQueueSend(qUpdQueue, &sender, 0);				// put in queue
		} else {
			if(TmrTime.second) {
				TmrFlags.Min1 = 0;							// ready for next minute
			}
		}
		
		// every 10 minutes save data to SD and Web
		if(((TmrTime.minute%10) == 0) && !TmrFlags.Min10) {	// every 10 minutes
			TmrFlags.Min10 = 1;								// indicated saved
			sender = UPD_S_10MTIMER;	
			xQueueSend(qUpdQueue, &sender, 0);				// put in queue
		} else {
			if(TmrTime.minute%10) {
				TmrFlags.Min10 = 0;							// ready for next 10 minutes
			}
		}
	}
} 

//********************************************************/
// TmrIncClock
// Adjust clock by one second
// 
// Parameters:
// TPtr		: pointer ti actual time structure
//********************************************************/
void TmrIncClock(volatile TmrTime_s *TPtr)
{
	uint8_t month;
	
	TPtr->second++;											// increment second and do the necessary adjustments
	if(TPtr->second == 60){									// check overflow
		TPtr->second = 0;									// adjust
		TPtr->minute++;
	}
	if(TPtr->minute == 60){									// again overflow handling
		TPtr->minute = 0;
		TPtr->hour++;
		if(TPtr->hour == 24){
			TPtr->hour = 0;
			TPtr->day++;
			TPtr->wday++;
			if(TPtr->wday == 8)
				TPtr->wday = 1;								// 0 indicated : no info. so start with 1
			month = TPtr->month;
			if((month == 2) && ((TPtr->year & 3) == 0))		// leap year
				month = 0;
			if(TPtr->day > MDays[month]) {
				TPtr->day = 1;
				TPtr->month++;
				if(TPtr->month == 13) {
					TPtr->month = 1;
					TPtr->year++;
					if(TPtr->year == 100)
						TPtr->year = 0;						// next century
				}
			}
		}
    }
}


