//*****************************************************************************
//
// File Name	: 'UpdateTask.c'
// Title		: Main function for Update actions
// Author		: Michael Meyer
// Created		: 17.08.2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: Tmr
//
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 
#include "Headers.h"


/** D E C L A R A T I O N S **************************************************/ 

#define UPD_STACK_SIZE		370								// adjust according value during runtime
#define	UPD_QUEUE_SIZE		100								// 100 events can be queued
#define UPD_BUF_SIZE		250								// size of general purpose buffer
#define UPD_DBERR_RTY       (5000 / portTICK_RATE_MS)		// 5 seconds retry delay if database update fails

// Status bits for UpdCurEvents
// These status bits are less time critical 
// Therefore, these status bits are delivered to the database together with the temperatur data every 10 minutes.

// Event bits for UpdCurEvents
// These event bits are more time critical 
// Therefore, these status bits are delivered to the database whenever they change but every 1 minute maximum
// Should be sufficient for longterm storage, while still saving memory space in data base (max 1440 entries per day)
#define	UPD_ST_SUMMER		(((uint32_t)1)<<0)				// 1 Sommerbetrieb, 0 Heizbetrieb
#define	UPD_ST_DAY			(((uint32_t)1)<<1)				// 1 Tagbetrieb, 0 Nachtbetrieb
#define	UPD_ST_HOL			(((uint32_t)1)<<2)				// 1 Ferienbetrieb, 0 kein Ferienbetrieb
#define	UPD_ST_AUT			(((uint32_t)1)<<3)				// 1 Automatikbetrieb, 0 Manuell
#define	UPD_EV_BRENNERON	(((uint32_t)1)<<4)				// 1 Brenner running
#define	UPD_EV_VPUMP		(((uint32_t)1)<<5)				// 1 Vorlaufpumpe is running
#define	UPD_EV_LPUMP		(((uint32_t)1)<<6)				// 1 Ladepumpe is running
#define	UPD_EV_CPUMP		(((uint32_t)1)<<7)				// 1 Zirkulationspumpe is running
#define	UPD_EV_WWPRIO		(((uint32_t)1)<<8)				// 1 Warmwasservorrang
#define	UPD_EV_LOADING		(((uint32_t)1)<<9)				// 1 WW Loading
#define	UPD_EV_RLOADING		(((uint32_t)1)<<10)				// 1 WW Nachladung
#define	UPD_EV_TESTGAS		(((uint32_t)1)<<11)				// 1 Abgastest ongoing
#define	UPD_EV_DIF			(((uint32_t)1)<<12)				// 1 Desinfection ongoing
#define	UPD_EV_ERR_FB		(((uint32_t)1)<<13)				// 1 Error: remote control
#define	UPD_EV_ERR_VLF		(((uint32_t)1)<<14)				// 1 Error: Vorlauffhler
#define	UPD_EV_ERR_EDIF		(((uint32_t)1)<<15)				// 1 Error: Desinfection
#define	UPD_EV_ERR_WWF		(((uint32_t)1)<<16)				// 1 Error: Warmwasserfhler
#define	UPD_EV_ERR_WWC		(((uint32_t)1)<<17)				// 1 Error: Warmwasser bleibt kalt
#define	UPD_EV_ERR_AND		(((uint32_t)1)<<18)				// 1 Error: Anodenfehler
#define	UPD_EV_ERR_BRN		(((uint32_t)1)<<19)				// 1 Error: Brennerstrung
#define	UPD_EV_ERR_BRNC		(((uint32_t)1)<<20)				// 1 Error: Kessel bleibt kalt
#define	UPD_EV_ERR_KSF		(((uint32_t)1)<<21)				// 1 Error: Kesselfhler
#define	UPD_EV_ERR_GASF		(((uint32_t)1)<<22)				// 1 Error: Abgasfhler
#define	UPD_EV_ERR_GASL		(((uint32_t)1)<<23)				// 1 Error: Abgas ber Limit

/** V A R I A B L E S ********************************************************/ 
xTaskHandle 				xUpdTaskHandle;					// task handle to do some debugging
xQueueHandle 				qUpdQueue;						// the handle to the Control queue
Flags_s						UpdFlags;						// some flags...
uint32_t					UpdCurEvents;					// current status of events
uint32_t					UpdLastEvents;					// nt status of events

/** Global local V A R I A B L E S *******************************************/ 
struct fat_fs_struct		*FileSystem;					// the selected file system
struct partition_struct		*Partition;						// the selected partition
struct fat_file_struct		*DbgFile;						// the file to log all events (debug)
struct fat_file_struct		*DBDataFile;					// the file to store HeizT Database data 
struct fat_file_struct		*DBEventFile;					// the file to store HeizE Database data 
uint8_t						UpdFileBuf[UPD_BUF_SIZE];		// general purpose buffer

// I hate this gcc PROGMEM handling just to save a few bytes of RAM...
const char WD0[]   PROGMEM = "--";
const char WD1[]   PROGMEM = "Mo";
const char WD2[]   PROGMEM = "Di";
const char WD3[]   PROGMEM = "Mi";
const char WD4[]   PROGMEM = "Do";
const char WD5[]   PROGMEM = "Fr";
const char WD6[]   PROGMEM = "Sa";
const char WD7[]   PROGMEM = "So";
const char PROGMEM *WTage[]  = {WD0, WD1, WD2, WD3, WD4, WD5, WD6, WD7};

#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
extern size_t 				xNextFreeByte;					// some debug variables
unsigned portBASE_TYPE 		uxHWMTmr, uxHWMKm, uxHWMUpd;	// to store stack size during runtime
#endif

/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
portTASK_FUNCTION_PROTO(UpdateTask, pvParameters);	
uint8_t UpdateLogFile(uint8_t index);
int16_t SendATCmd(char *Buf);
void	UpdateDateTime(void);
void	UpdateDataBase(void);
void	UpdateEvDataBase(void);
void	UpdateDebug(void);
void	UpdateDayValues(void);
void	UpdateFMValues(uint8_t index);
void	CheckAction(void);
void	CheckNewEvents();
void 	GenLogEntry(uint8_t index);
void 	UpdateWebSite(uint8_t index);
uint8_t	CheckDateTime(void);
uint8_t	CreateSDFiles(void);
uint8_t UpdateHeizTFile(void);
uint8_t UpdateHeizEFile(void);

/** C O D E ******************************************************************/ 


//********************************************************/
// UpdTaskInit
// Initialisation of Update Task, called once from main
// 
// Parameters:
// none
//********************************************************/
uint8_t UpdTaskInit(void)
{
	portENTER_CRITICAL();
	{
		qUpdQueue = xQueueCreate(UPD_QUEUE_SIZE, sizeof(uint8_t));		// create Update queue
		xTaskCreate(UpdateTask, (signed portCHAR *)"U", UPD_STACK_SIZE, NULL, UPD_UPD_PRIO, &xUpdTaskHandle); // create Update task	
	}
	portEXIT_CRITICAL();

	return TRUE; 
}

//********************************************************/
// UpdateTask
// Handles Update activities
// 
// Parameters:
// none
//********************************************************/
portTASK_FUNCTION(UpdateTask, pvParameters)
{
 	(void)			pvParameters;							// Just to stop compiler warnings.
	uint8_t			cmd;									// received command

	UpdFlags.TimeValid = 0;									// Start with invalid time
	UpdFlags.StoreEvent = 1;								// Start clean and send events after reset
	DBDataFile = 0;											// this is also used to indicate already opened file
	UpdCurEvents = 0;										// reset current status/events

	while(1) {
		xQueueReceive(qUpdQueue, &cmd, portMAX_DELAY);		// wait for something to do
		if(!UpdFlags.TimeValid) {
			if(cmd == UPD_S_10STIMER) {
				UpdFlags.TimeValid = CheckDateTime();		// get Date/Time Strings from WLAN Module
				CreateSDFiles();							// create and open files on SD card, if necessary
			}
		} else {
			switch(cmd) {
				case UPD_S_10STIMER:						// wake up from timer every 10 seconds
					UpdateDateTime();						// update date and time on web site
					UpdateDebug();							// update debug values
					CheckAction();							// check for any action to be carried out
					break;
				case UPD_S_1MTIMER:							// wake up from timer every minute
					CheckNewEvents();						// check for new events, update DB and SD card backup if necessary
					break;
				case UPD_S_10MTIMER:						// wake up from timer every 10 minutes
					UpdateDataBase();						// update web data base
					UpdateHeizTFile();						// entry into SD card backup
					break;
				case UPD_S_SYNC_TIMER:						// wake up, if it is time for sync
					CheckDateTime();						// try to re-sync. Do not stop if fails
					break;
				case UPD_S_DAY_TIMER:						// wake up, if new day
					UpdateDayValues();						// everything, what needs handling at 00:00:00
					break;
				default:
					UpdateFMValues(cmd);					// check, if we need to update some FRAM values				
					UpdateLogFile(cmd);						// write log entry into log file on SD-card
					UpdateWebSite(cmd);						// update value in website of WLAN module
					break;
	
			}
		}
	}	
} 

/********************************************************/
// CheckAction
// Checks for an action string in UF12 and carries out action if found
// 
// Parameters:
// none
//********************************************************/
void CheckAction(void)
{
	uint8_t	updated=0;

	WLANSendCom(PSTR("AT+iUF12?\r"), 1);					// retrieve UF12
	if(WLANGetResp((char *)UpdFileBuf, UPD_BUF_SIZE) < 0) {	// from WLAN module
		return;												// error -> ignore response
	}		

	// check for actions and carry them out
	if(!strncmp_P((char *)UpdFileBuf, PSTR("MINMAX"), 6)) {	// reset of min max values
		// not used, yet
		updated = 1;
	} else
	if(!strncmp_P((char *)UpdFileBuf, PSTR("CLEAR"), 5)) {	// reset all FRAM
		FRAMClear();										// clear all FRAM structures
		updated = 1;
	}

	if(updated) {											// give some indication back to user
		WLANSendCom(PSTR("AT+iUF12=\"Done\"\r"), 1);
		WLANGetResp((char *)UpdFileBuf, UPD_BUF_SIZE);
	}
}


/********************************************************/
// DayOfWeek
// Calculates the weekday
// 0 = Sonntag
// 1 <= m <= 12
// y > 1752 oder so
// Parameters:
// none
//********************************************************/
uint8_t DayOfWeek(uint16_t y, uint16_t m, uint16_t d)      
{
	static int t[] = {0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4};

	y -= m < 3;
	return (y + y/4 - y/100 + y/400 + t[m-1] + d) % 7;
}

/********************************************************/
// CheckDateTime
// Verifies if date and time are valid
// set Date Time strings if yes
// 
// Parameters:
// none
// Returns:		0 for Invalid time
//				1 for Valid Time
//********************************************************/
uint8_t CheckDateTime(void)
{
	uint8_t	sec,min,hour,wday;

	// get current date/time
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iRP8\r"));
	if(SendATCmd((char *)UpdFileBuf) < 0) {
		LED3_OFF;											// indicate loss of time sync
		return 0;	
	}

	if(UpdFileBuf[0]=='2') {
		LED3_ON;											// indicate readiness, time and date are set from internet
		hour = atoi(((char *)UpdFileBuf)+11);				// hour
		min = atoi(((char *)UpdFileBuf)+14);				// minutes
		sec = atoi(((char *)UpdFileBuf)+17);				// seconds
		cli();
		TmrTime.hour = hour;
		TmrTime.minute = min;
		TmrTime.second = sec;
		sei();

		hour = atoi(((char *)UpdFileBuf)+2);				// year
		min = atoi(((char *)UpdFileBuf)+5);					// month
		sec = atoi(((char *)UpdFileBuf)+8);					// day
		
		wday = DayOfWeek(2000+hour, min, sec);				// calculate weekday
		if(wday==0) {
			wday = 7;										// 0 is unused, Monday = 1, Tuesday = 2, ... Sunday = 8;
		}
		cli();
		TmrTime.year = hour;
		TmrTime.month = min;
		TmrTime.day = sec;
		TmrTime.wday = wday;								// day of week
		sei();
		return 1;
	} else {
		LED3_OFF;											// indicate loss of time sync
		return 0;
	}
}

/********************************************************/
// UpdUpdateDateTime
// Stores actual date/time to website
// 
// Parameters:
// none
//********************************************************/
void UpdateDateTime(void)
{
	char	WDay[3];

	strcpy_P(WDay, (PGM_P)pgm_read_word(&WTage[TmrTime.wday]));
	// update date and time on web site
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHDATE=\"%s, %02u.%02u.20%02u\"\r"), WDay,
														TmrTime.day, TmrTime.month, TmrTime.year);
	SendATCmd((char *)UpdFileBuf);
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHTIME=\"%02u:%02u:%02u\"\r"), 
														TmrTime.hour, TmrTime.minute, TmrTime.second);
	SendATCmd((char *)UpdFileBuf);
}

/********************************************************/
// CreateSDFiles
// 
// Parameters:
// none
// Returns:		0 for ERROR
//				1 for OK
//********************************************************/
uint8_t CreateSDFiles(void)
{
	struct fat_dir_entry_struct	dir_entry;
	struct fat_dir_struct		*dir;
	
	if(!DBDataFile) {										// is the file on card already open, no -> open card
		/* setup sd card slot */
		KM_WAIT_SPI();
		if(!sd_raw_init()) {								// try to open the physical interface
			KM_RELEASE_SPI();
			WLANSpiInit();									// make sure to switch SPI interface to fast mode
			return 0;										// no success! return with error
		}
		KM_RELEASE_SPI();
		KM_WAIT_SPI();
		Partition = partition_open(							//  open first partition
			sd_raw_read,
			sd_raw_read_interval,
			sd_raw_write,
			sd_raw_write_interval,
			0);

		// If the partition did not open, assume the storage device
		// is a "superfloppy", i.e. has no MBR.
		if(!Partition) {									// so try again
			Partition = partition_open(
				sd_raw_read,
				sd_raw_read_interval,
				sd_raw_write,
				sd_raw_write_interval,
				-1);
		}
		KM_RELEASE_SPI();
		if(!Partition) {									// still no partition found
			return 0;										// give up
		}
	     
		KM_WAIT_SPI();
		FileSystem = fat_open(Partition);					// open file system
		KM_RELEASE_SPI();
		if(!FileSystem) {									// oops, no file system
			return 0;										// cannot continue
		}

		strcpy_P((char *)UpdFileBuf, PSTR("\\"));			// go to root
		KM_WAIT_SPI();
		if(fat_get_dir_entry_of_path(FileSystem, (char *)UpdFileBuf, &dir_entry) == 0) {
			KM_RELEASE_SPI();
			return 0;
		}
		KM_RELEASE_SPI();
	
		
		KM_WAIT_SPI();
		dir = fat_open_dir(FileSystem, &dir_entry);			// open root
		KM_RELEASE_SPI();
		if(!dir)
			return 0;

		// create HeizT file
		strcpy_P((char *)UpdFileBuf, PSTR("HeizT.csv"));	// create the data file if necessary
		KM_WAIT_SPI();
		fat_create_file(dir, (char *)UpdFileBuf, &dir_entry);
		
		if(!(DBDataFile = fat_open_file(FileSystem, &dir_entry))) {	// open the Log File	
			KM_RELEASE_SPI();
			return 0;
		}		
		KM_RELEASE_SPI();
		
		// create HeizE file
		strcpy_P((char *)UpdFileBuf, PSTR("HeizE.csv"));	// create the data file if necessary
		KM_WAIT_SPI();
		fat_create_file(dir, (char *)UpdFileBuf, &dir_entry);
		
		if(! (DBEventFile = fat_open_file(FileSystem, &dir_entry))) {	// open the Log File	
			KM_RELEASE_SPI();
			return 0;
		}		
		KM_RELEASE_SPI();		
		
		// create Debug file
		strcpy_P((char *)UpdFileBuf, PSTR("HeizDebug.txt"));	// create the data file if necessary
		KM_WAIT_SPI();
		fat_create_file(dir, (char *)UpdFileBuf, &dir_entry);
		
		if(! (DbgFile = fat_open_file(FileSystem, &dir_entry))) {	// open the Log File	
			KM_RELEASE_SPI();
			return 0;
		}		
		KM_RELEASE_SPI();		
		
		KM_WAIT_SPI();										// make sure data is written physically to card
		sd_raw_sync();
		KM_RELEASE_SPI();
	}
	return 1;
}


/********************************************************/
// UpdateLogFile
// 
// Parameters:
// none
// Returns:		0 for ERROR
//				1 for OK
//********************************************************/
uint8_t UpdateLogFile(uint8_t index)
{
	int32_t						offset;
	
	KM_WAIT_SPI();
	offset = 0l;											// go to end of file
	fat_seek_file(DbgFile, &offset, FAT_SEEK_END);
	KM_RELEASE_SPI();

	GenLogEntry(index);										// retrieve line of logging data to store

	KM_WAIT_SPI();											// write now
	if(fat_write_file(DbgFile, UpdFileBuf, strlen((char *)UpdFileBuf)) != strlen((char *)UpdFileBuf)) {
		KM_RELEASE_SPI();
		return 0;
	}
	KM_RELEASE_SPI();

	KM_WAIT_SPI();											// make sure data is written physically to card
	sd_raw_sync();
	KM_RELEASE_SPI();
	return 1;												// all fine
}

/********************************************************/
// UpdateHeizTFile
// 
// Parameters:
// none
// Returns:		0 for ERROR
//				1 for OK
//********************************************************/
uint8_t UpdateHeizTFile(void)
{
	int32_t						offset;
	
	sprintf_P((char *)UpdFileBuf,PSTR(";20%02u-%02u-%02u;%02u:%01u0:00;"), 	TmrTime.year, TmrTime.month, TmrTime.day,
																			TmrTime.hour, TmrTime.minute/10);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u.%u;"),KmParamD[KM_I_TIR]/2, (KmParamD[KM_I_TIR]%2)?5:0);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u.%u;"),KmParamD[KM_I_TSR]/2, (KmParamD[KM_I_TSR]%2)?5:0);
	if(KmParamD[KM_I_TIA]>128) {
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("-%u;"),KmParamD[KM_I_TIA]-128);
	} else {
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIA]);
	}
	if(KmParamD[KM_I_TVA]>128) {
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("-%u;"),KmParamD[KM_I_TVA]-128);
	} else {
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TVA]);
	}
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIKV]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TSKV]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TBRE]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TBRA]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIV]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TSV]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIWW]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TSWW]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),(uint8_t)FRAMReadLong(FM_MAX_ABGAS));
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIHKP10]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIHK0]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%u;"),KmParamD[KM_I_TIHKM10]);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%lu\r\n"),UpdCurEvents);
	
	KM_WAIT_SPI();
	offset = 0l;											// go to end of file
	fat_seek_file(DBDataFile, &offset, FAT_SEEK_END);
	KM_RELEASE_SPI();
	
	KM_WAIT_SPI();											// write now
	if(fat_write_file(DBDataFile, UpdFileBuf, strlen((char *)UpdFileBuf)) != strlen((char *)UpdFileBuf)) {
		KM_RELEASE_SPI();
		return 0;
	}
	KM_RELEASE_SPI();

	KM_WAIT_SPI();											// make sure data is written physically to card
	sd_raw_sync();
	KM_RELEASE_SPI();
	return 1;												// all fine
}

/********************************************************/
// UpdateHeizEFile
// 
// Parameters:
// none
// Returns:		0 for ERROR
//				1 for OK
//********************************************************/
uint8_t UpdateHeizEFile(void)
{
	int32_t						offset;
	
	sprintf_P((char *)UpdFileBuf,PSTR(";20%02u-%02u-%02u;%02u:%02u:00;"), 	TmrTime.year, TmrTime.month, TmrTime.day,
																			TmrTime.hour, TmrTime.minute);
	sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("%lu\r\n"),UpdCurEvents);

	KM_WAIT_SPI();
	offset = 0l;											// go to end of file
	fat_seek_file(DBDataFile, &offset, FAT_SEEK_END);
	KM_RELEASE_SPI();
	
	KM_WAIT_SPI();											// write now
	if(fat_write_file(DBEventFile, UpdFileBuf, strlen((char *)UpdFileBuf)) != strlen((char *)UpdFileBuf)) {
		KM_RELEASE_SPI();
		return 0;
	}
	KM_RELEASE_SPI();

	KM_WAIT_SPI();											// make sure data is written physically to card
	sd_raw_sync();
	KM_RELEASE_SPI();
	return 1;												// all fine
}


/********************************************************/
// GenLogEntry
// generates Log Entry string in UpdFileBuf
// 
// Parameters:
// Index of parameter
//********************************************************/
void GenLogEntry(uint8_t index)
{
	sprintf_P((char *)UpdFileBuf,PSTR("20%02u-%02u-%02u %02u:%02u:%02u\t"),
	TmrTime.year, TmrTime.month, TmrTime.day, TmrTime.hour, TmrTime.minute, TmrTime.second);

	switch(pgm_read_byte(&KmParamR[index].ParType)) {
		case KM_PAR_TEMP:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("%uC"), KmParamD[index]);
			break;
		case KM_PAR_MINUTES:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("%u min"), KmParamD[index]);
			break;
		case KM_PAR_NUMBER:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("%u"), KmParamD[index]);
			break;
		case KM_PAR_PERCENT:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("%u%%"), KmParamD[index]);
			break;
		case KM_PAR_TEMP_HALF:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("%u.%uC"), KmParamD[index]/2, (KmParamD[index]%2)?5:0);
			break;
		case KM_PAR_TEMP_NEG:
			if(KmParamD[index]>128) {
				sprintf_P((char *)(UpdFileBuf+100), PSTR("-%uC"), KmParamD[index]-128);
			} else {
				sprintf_P((char *)(UpdFileBuf+100), PSTR("%uC"), KmParamD[index]);
			}
			break;
		case KM_PAR_BITFIELD:
			switch(pgm_read_byte(&KmParamR[index].Id)) {
				default:									// no specific decoding
					sprintf_P((char *)(UpdFileBuf+100), PSTR("%02X"), KmParamD[index]);
					break;
			}
			break;
	}
	memcpy(UpdFileBuf+strlen((char *)UpdFileBuf),UpdFileBuf+100, strlen((char *)(UpdFileBuf+100))+1);
	strcat_P((char *)UpdFileBuf, PSTR("\t\t"));
	strcat_P((char *)UpdFileBuf, (const char *)pgm_read_word(&KmParamR[index].Name));
	strcat_P((char *)UpdFileBuf, PSTR("\r\n"));
}		


/********************************************************/
// UpdateWebSite
// Update the changed value on the Webserver of the WLAN Module
// 
// Parameters:
// Index of parameter
//********************************************************/
void UpdateWebSite(uint8_t index)
{

	strcpy_P((char *)UpdFileBuf, PSTR("AT+i"));
	strcat_P((char *)UpdFileBuf,(const char *)pgm_read_word(&KmParamR[index].ATiName));
	
	switch(pgm_read_byte(&KmParamR[index].ParType)) {
		case KM_PAR_TEMP:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u\"\r"), KmParamD[index]);
			break;
		case KM_PAR_MINUTES:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u\"\r"), KmParamD[index]);
			break;
		case KM_PAR_NUMBER:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u\"\r"), KmParamD[index]);
			break;
		case KM_PAR_PERCENT:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u\"\r"), KmParamD[index]);
			break;
		case KM_PAR_TEMP_HALF:
			sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u.%u\"\r"), KmParamD[index]/2, (KmParamD[index]%2)?5:0);
			break;
		case KM_PAR_TEMP_NEG:
			if(KmParamD[index]>128) {
				sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"-%u\"\r"), KmParamD[index]-128);
			} else {
				sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%u\"\r"), KmParamD[index]);
			}
			break;
		case KM_PAR_BITFIELD:
			switch(pgm_read_byte(&KmParamR[index].Id)) {
				default:									// no specific decoding
					sprintf_P((char *)(UpdFileBuf+100), PSTR("=\"%02X\"\r"), KmParamD[index]);
					break;
			}
			break;
	}
	// concat strings
	memcpy(UpdFileBuf+strlen((char *)UpdFileBuf),UpdFileBuf+100, strlen((char *)(UpdFileBuf+100))+1);
	SendATCmd((char *)UpdFileBuf);

	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHEVST=\"%lu\"\r"), UpdCurEvents);	// Update events on web site
	SendATCmd((char *)UpdFileBuf);
}		

/********************************************************/
// UpdateDataBase
// Stores actual data to database
// 
// Parameters:
// none
//********************************************************/
void UpdateDataBase(void)
{
	uint8_t	retry;

	for(retry=0;retry<3;retry++) {
		// write data to internet data base
		sprintf_P((char *)UpdFileBuf,PSTR("AT+iSLNK:D=20%02u-%02u-%02u&Z=%02u:%01u0:00"), 	TmrTime.year, TmrTime.month, TmrTime.day,
																							TmrTime.hour, TmrTime.minute/10);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIR=%u.%u"),KmParamD[KM_I_TIR]/2, (KmParamD[KM_I_TIR]%2)?5:0);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TSR=%u.%u"),KmParamD[KM_I_TSR]/2, (KmParamD[KM_I_TSR]%2)?5:0);
		if(KmParamD[KM_I_TIA]>128) {
			sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIA=-%u"),KmParamD[KM_I_TIA]-128);
		} else {
			sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIA=%u"),KmParamD[KM_I_TIA]);
		}
		if(KmParamD[KM_I_TVA]>128) {
			sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TVA=-%u"),KmParamD[KM_I_TVA]-128);
		} else {
			sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TVA=%u"),KmParamD[KM_I_TVA]);
		}
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIKV=%u"),KmParamD[KM_I_TIKV]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TSKV=%u"),KmParamD[KM_I_TSKV]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TBRE=%u"),KmParamD[KM_I_TBRE]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TBRA=%u"),KmParamD[KM_I_TBRA]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIV=%u"),KmParamD[KM_I_TIV]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TSV=%u"),KmParamD[KM_I_TSV]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIWW=%u"),KmParamD[KM_I_TIWW]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TSWW=%u"),KmParamD[KM_I_TSWW]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIABG=%u"),(uint8_t)FRAMReadLong(FM_MAX_ABGAS));
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIHKP10=%u"),KmParamD[KM_I_TIHKP10]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIHK0=%u"),KmParamD[KM_I_TIHK0]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&TIHKM10=%u"),KmParamD[KM_I_TIHKM10]);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&ST=%lu\r\n.\r\n"),UpdCurEvents);
		WLANSendCom((char *)UpdFileBuf, 0);
		memset(UpdFileBuf,0,30);												// just something else than O
		if(WLANGetSLNK((char *)UpdFileBuf, UPD_BUF_SIZE) >= 0) {				// if this is <0 then we have an error
			if(!strncmp_P((char *)&UpdFileBuf[11], PSTR("OK_DB"),5)) {			// OK_DB is generated by receiving PHP file
				break;
			}
		}
		vTaskDelay(UPD_DBERR_RTY);												// wait a bit, then retry
	}
}

/********************************************************/
// UpdateEvDataBase
// Stores actual events to database on internet
// 
// Parameters:
// none
//********************************************************/
void UpdateEvDataBase(void)
{
	uint8_t	retry;

	for(retry=0;retry<3;retry++) {
		// write data to internet data base
		sprintf_P((char *)UpdFileBuf,PSTR("AT+iSLNK:D=20%02u-%02u-%02u&Z=%02u:%02u:00"), 	TmrTime.year, TmrTime.month, TmrTime.day,
																							TmrTime.hour, TmrTime.minute);
		sprintf_P((char *)&UpdFileBuf[strlen((char *)UpdFileBuf)], PSTR("&EV=%lu\r\n.\r\n"),UpdCurEvents);
		WLANSendCom((char *)UpdFileBuf, 0);
		memset(UpdFileBuf,0,30);												// just something else than O
		if(WLANGetSLNK((char *)UpdFileBuf, UPD_BUF_SIZE) >= 0) {				// if this is <0 then we have an error
			if(!strncmp_P((char *)&UpdFileBuf[11], PSTR("OK_DB"),5)) {			// OK_DB is generated by receiving PHP file
				break;
			}
		}
		vTaskDelay(UPD_DBERR_RTY);												// wait a bit, then retry
	}
}


/********************************************************/
// UpdateFMValue
// Check if FM values need to be adjusted
// 
// Parameters:
// none
//********************************************************/
void UpdateFMValues(uint8_t index)
{
	switch(index) {
		case KM_I_TIABG:
			if(KmParamD[index] > (uint8_t)FRAMReadLong(FM_MAX_ABGAS)) {	// collect highest Abgas temperature
				FRAMWriteLong((uint32_t)KmParamD[index], FM_MAX_ABGAS);	// new higher value -> store
			}
			break;
		case KM_I_HKBW1:
			if(KmParamD[index] & KM_HKBW1_AUT) {			// Automatikbetrieb
				UpdCurEvents |= UPD_ST_AUT;	
			} else {
				UpdCurEvents &= ~UPD_ST_AUT;	
			}
			if(KmParamD[index] & KM_HKBW1_HOL) {			// holidays
				UpdCurEvents |= UPD_ST_HOL;	
			} else {
				UpdCurEvents &= ~UPD_ST_HOL;	
			}
			if(KmParamD[index] & KM_HKBW1_WWPRIO) {			// Vorrang Warmwasser 
				UpdCurEvents |= UPD_EV_WWPRIO;	
			} else {
				UpdCurEvents &= ~UPD_EV_WWPRIO;	
			}
			break;
		case KM_I_HKBW2:
			if(KmParamD[KM_I_HKBW2] & KM_HKBW2_SUMMER) {
				UpdCurEvents |= UPD_ST_SUMMER;				// Sommerbetrieb
			} else {
				UpdCurEvents &= ~UPD_ST_SUMMER;	
			}
			if(KmParamD[KM_I_HKBW2] & KM_HKBW2_DAY) {
				UpdCurEvents |= UPD_ST_DAY;					// Tagbetrieb
			} else {
				UpdCurEvents &= ~UPD_ST_DAY;	
			}
			if(KmParamD[index] & KM_HKBW2_FB) {				// Keine Kommunikation mit FB
				UpdCurEvents |= UPD_EV_ERR_FB;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_FB;	
			}
			if(KmParamD[index] & KM_HKBW2_VLF) {			// Problem Vorlauffhler
				UpdCurEvents |= UPD_EV_ERR_VLF;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_VLF;	
			}
			break;
		case KM_I_HWWBW1:
			if(KmParamD[index] & KM_HWWBW1_DIF) {			// Desinfektion luft
				UpdCurEvents |= UPD_EV_DIF;	
			} else {
				UpdCurEvents &= ~UPD_EV_DIF;	
			}
			if(KmParamD[index] & KM_HWWBW1_RLOADING) {		// Nachladung
				UpdCurEvents |= UPD_EV_RLOADING;	
			} else {
				UpdCurEvents &= ~UPD_EV_RLOADING;	
			}
			if(KmParamD[index] & KM_HWWBW1_EDIF) {			// Fehler Desinfektion
				UpdCurEvents |= UPD_EV_ERR_EDIF;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_EDIF;	
			}
			if(KmParamD[index] & KM_HWWBW1_WWF) {			// Fehler Warmwasserfhler
				UpdCurEvents |= UPD_EV_ERR_WWF;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_WWF;	
			}
			if(KmParamD[index] & KM_HWWBW1_WWC) {			// Fehler Warmwasser bleibt kalt
				UpdCurEvents |= UPD_EV_ERR_WWC;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_WWC;	
			}
			if(KmParamD[index] & KM_HWWBW1_AND) {			// Fehler Anode
				UpdCurEvents |= UPD_EV_ERR_AND;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_AND;	
			}
			break;
		case KM_I_HWWBW2:
			if(KmParamD[index] & KM_HWWBW2_LOADING) {		// Warmwasser Loading
				UpdCurEvents |= UPD_EV_LOADING;	
			} else {
				UpdCurEvents &= ~UPD_EV_LOADING;	
			}
			break;
		case KM_I_HKFEH:
			if(KmParamD[index] & KM_HKFEH_BRN) {			// Brennerstrung
				UpdCurEvents |= UPD_EV_ERR_BRN;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_BRN;	
			}
			if(KmParamD[index] & KM_HKFEH_KSF) {			// Kesselfhler
				UpdCurEvents |= UPD_EV_ERR_KSF;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_KSF;	
			}
			if(KmParamD[index] & KM_HKFEH_BRNC) {			// Kessel bleibt kalt
				UpdCurEvents |= UPD_EV_ERR_BRNC;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_BRNC;	
			}
			if(KmParamD[index] & KM_HKFEH_GASF) {			// Abgasfhler
				UpdCurEvents |= UPD_EV_ERR_GASF;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_GASF;	
			}
			if(KmParamD[index] & KM_HKFEH_GASL) {			// Abgas ber Grenzwert
				UpdCurEvents |= UPD_EV_ERR_GASL;	
			} else {
				UpdCurEvents &= ~UPD_EV_ERR_GASL;	
			}
			break;
		case KM_I_HBREN:
			if(KmParamD[index] & KM_HBREN_BRENNERON) {		// Brenner an
				UpdCurEvents |= UPD_EV_BRENNERON;
			} else {
				UpdCurEvents &= ~UPD_EV_BRENNERON;	
			}
			break;
		case KM_I_HPUMP:
			if(KmParamD[index] & KM_HPUMP_VPUMP) {			// Vorlaufpumpe (an ist 100 -> bit 5 gesetzt) 
				UpdCurEvents |= UPD_EV_VPUMP;
			} else {
				UpdCurEvents &= ~UPD_EV_VPUMP;	
			}
			break;
		case KM_I_HKBET:
			if(KmParamD[index] & KM_HKBET_TESTGAS) {		// Abgastest
				UpdCurEvents |= UPD_EV_TESTGAS;
			} else {
				UpdCurEvents &= ~UPD_EV_TESTGAS;	
			}
			break;
		case KM_I_HWWLP:
			if(KmParamD[index] & KM_HWWLP_LPUMP) {			// Warmwasser Ladepumpe
				UpdCurEvents |= UPD_EV_LPUMP;
			} else {
				UpdCurEvents &= ~UPD_EV_LPUMP;	
			}
			if(KmParamD[index] & KM_HWWLP_CPUMP) {			// Warmwasser Zirkulationspumpe
				UpdCurEvents |= UPD_EV_CPUMP;
			} else {
				UpdCurEvents &= ~UPD_EV_CPUMP;	
			}
			break;
	}
}


/********************************************************/
// UpdateDayValues
// All actions at 00:00:00 every day
// 
// Parameters:
// none
//********************************************************/
void UpdateDayValues(void)
{
	UpdFlags.StoreEvent = 1;								// Store event value in any case at 00:00 every day
	FRAMWriteLong(0l, FM_MAX_ABGAS);						// reset Abgas value every day
}



/********************************************************/
// CheckNewEvents
// Checks for new events every minute
// if new events are discovered, the events will be stored
// 
// Parameters:
// none
//********************************************************/
void CheckNewEvents(void)
{
	if(UpdFlags.StoreEvent || (UpdLastEvents != UpdCurEvents)) {
		UpdateEvDataBase();									// save events to web data base
		UpdateHeizEFile();									// entry into SD card backup
		UpdFlags.StoreEvent = 0;							// reset store event flag in any case, we have stored
	}
	UpdLastEvents = UpdCurEvents; 							// save current events for next time
}


/********************************************************/
// UpdDebug
// Update all debug values
// 
// Parameters:
// none
//********************************************************/
void UpdateDebug(void)
{
	#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)			// only, if we are in debug mode
	uint8_t	resets;
					
	// update debug values 
	uxHWMUpd = uxTaskGetStackHighWaterMark(NULL);
	uxHWMKm = uxTaskGetStackHighWaterMark(xKmTaskHandle);
	uxHWMTmr = uxTaskGetStackHighWaterMark(xTmrTaskHandle);
	// TimerTask Stack Watermark
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMTMR=\"%d\"\r"), uxHWMTmr );
	SendATCmd((char *)UpdFileBuf);
	// ISVTask Stack Watermark
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMKM=\"%d\"\r"), uxHWMKm );
	SendATCmd((char *)UpdFileBuf);
	// UpdTask Stack Watermark
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMUPD=\"%d\"\r"), uxHWMUpd );
	SendATCmd((char *)UpdFileBuf);
	// Heap Watermark
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMHEAP=\"%d\"\r"), xPortGetFreeHeapSize());
	SendATCmd((char *)UpdFileBuf);
	// resets
	portDISABLE_INTERRUPTS();
	resets = eeprom_read_byte(&ManResCnt);
	portENABLE_INTERRUPTS();
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMRES=\"%d\"\r"), resets);
	SendATCmd((char *)UpdFileBuf);
	// SPI Timeouts
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMSPI=\"%d\"\r"), WLANTmeOut);
	SendATCmd((char *)UpdFileBuf);
	// 3964 Protocol problems
	sprintf_P((char *)UpdFileBuf, PSTR("AT+iHHMRXE=\"%d\"\r"), KmRxErr);
	SendATCmd((char *)UpdFileBuf);
	#endif
}

/********************************************************/
// SendATCmd
// Completes and send the Update String
// 
// Parameters:
// none
//********************************************************/
int16_t SendATCmd(char *Buf)
{
	WLANSendCom(Buf, 0);									// send the AT command
	return (WLANGetResp(Buf, UPD_BUF_SIZE));				// get back the response
}

