//*****************************************************************************
//
// File Name	: 'Wlan.c'
// Title		: Provide WLAN functionality
// Author		: Michael Meyer
// Created		: 12.08.2010
// Version		: 0.1
// Target MCU	: AVR ATmega 644p
// Indicator	: WLAN
//
// 11.059200 MHz external clock
//
//*****************************************************************************

/** I N C L U D E S **********************************************************/ 

#include "Headers.h"


/** D E C L A R A T I O N S **************************************************/ 
#define	SPIR_TMEOUT_CON				(45000 / portTICK_RATE_MS)	// timeout for wait on interrupt, >35 seconds as that is the connection timeout
#define	SPIR_TMEOUT_AT				(300 / portTICK_RATE_MS)	// timeout for normal AT commands

#define	WLAN_PSE					(1800 / portTICK_RATE_MS)	// AT+iPSE = 2, so assume power save after 1,8 seconds

#define	SET_WLAN_SPI				{ SPCR |= (1<<CPHA); }		// SPI mode for WLAN module
#define	RESET_WLAN_SPI				{ SPCR &= ~(1<<CPHA);}		// SPI mode for the rest of SPI devices

#define	WLAN_MAX_TICK				65000						// max value for WLAN Tick

/** V A R I A B L E S ********************************************************/ 
xSemaphoreHandle 					qWLANMutex;					// the WLAN Mutex

#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
uint16_t volatile					WLANTmeOut;					// Number of timeouts in SPI handling
#endif

/** Global local V A R I A B L E S *******************************************/ 
static xSemaphoreHandle volatile	qSPIRSema;					// the SPIR Semaphore
static uint16_t volatile			WlanTick;					// the tick counter to check for WLAN Power Save

/** Local F U N C T T I O N  P R T O T Y P E S *******************************/ 
void WriteSPIInt(uint8_t Data);
uint16_t GetAndResWLANTick(void);
	
/** C O D E ******************************************************************/ 


//********************************************************/
// WLAN DTR Interrupt routine
// This interrupt is activated, when WLAN Module has new TCP data
// Related to WLAN_DATA_RDY
// not used, yet
// 
// Parameters:
// none
//********************************************************/
ISR (INT1_vect)  
{
	WLAN_DIS_DATA_INT;										// disable this interrupt
}


//********************************************************/
// WLAN SPI_RDY Interrupt routine
// This interrupt is activated, when WLAN Module 
// has response data available at SPI port
// 
// Parameters:
// none
//********************************************************/
ISR (INT0_vect)  
{
	WLAN_DIS_SPIR_INT;										// disable this interrupt

	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE; 
	
	xSemaphoreGiveFromISR(qSPIRSema, &xHigherPriorityTaskWoken); // indicate occurence of interrupt
	if(xHigherPriorityTaskWoken == pdTRUE) {
		taskYIELD();										// wakeup waiting task, if higher prio
    }
}

//********************************************************/
// vApplicationTickHook
// This interrupt is activated from RTOS System Tick
// Used to check if WLAN is already in Power-Save mode
// Is called every 5ms in this configuration
// (configTICK_RATE_HZ == 200)
//
// PSE functionality currently disabled!!
// 
// Parameters:
// none
//********************************************************/
void vApplicationTickHook(void) {
	if(WlanTick < WLAN_MAX_TICK) {								// limit is 65000*0,005 sec = 325 sec
		WlanTick = 0;					// ++;					// no wrap!
	}
}

//********************************************************/
// GetAndResWLANTick
// Reset the WLAN Tick counter and returns the number of 
// ticks since last call
// 
// Parameters:
// none
//********************************************************/
uint16_t GetAndResWLANTick(void) {
	uint16_t	ret;

	taskDISABLE_INTERRUPTS();								// modify WlanTick during disabled interrupts only!
	ret = WlanTick;											// retrieve current tick value
	WlanTick = 0;											// rest for next measurement
	taskENABLE_INTERRUPTS();								// allow interrupts again
	return ret;												// return number of ticks since last call
}

//********************************************************/
// WLANInit
// Initialisation of WLAN functionality
// 
// Parameters:
// none
//********************************************************/
void WLANInit(void)
{
	vSemaphoreCreateBinary(qSPIRSema);						// create semaphore to synchronise SPI READY interrupt
	qWLANMutex = xSemaphoreCreateMutex();					// create Mutex in order to synchronise access to WLAN interface
															// WLAN Module requires a clean command/response sequence....
	#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)	
	WLANTmeOut	= 0;										// Debugging SPI timeouts... happens from time to time
	#endif

	WLANRES_PORT &=	~WLAN_RES;								// put RESET to active
	KM_DELAY_MS(200);										// give module som time to reset
	WLANRES_PORT |=	WLAN_RES;								// put RESET to inactive
	while(WLANICHIP_PIN & WLAN_ICHIP_RDY);					// wait until WLAN is ready to receive AT commands (about 1.8 seconds!)
	KM_DELAY_MS(100);										// just give it a bit more time...
	WLAN_RELEASE();											// ensure WLAN is accessible
	
	taskDISABLE_INTERRUPTS();								// modify WlanTick during disabled interrupts only!
	WlanTick = WLAN_MAX_TICK;								// set WLAN Tick to max value -> assume WLAN entered powersave after reset 
	taskENABLE_INTERRUPTS();								// allow interrupts again
}


//********************************************************/
// WLANSendCom
// Writes a command to the WLAN Module, 0-terminated string!
// 
// Parameters: 	Buffer with command
//				Zero terminated, in RAM of Flash buffer!
// Returns:		none
//********************************************************/
void WLANSendCom(char *Com, uint8_t	InFlash)
{
	uint8_t	wait, first;

	WLAN_BLOCK();											// wait for WLAN interface available and block it for now
	KM_WAIT_SPI();											// wait for SPI bus available
	SET_WLAN_SPI;											// send command
	WLANCS_PORT &= ~WLAN_CS;								// activate WLAN SPI interface

	// WLAN module has problems to wake up from powersave quickly enough when woken by SPI activity.
	// The first bytes of first block sent to the WLAN Module after it entered power save
	// must be sent slowly (delay between CS and first and some consecutive bytes).
	// To maintain a good performance for the communication with the module, we
	// check if the module has entered powersave and insert the delays only under this condition.
	// Consecutive commands are send without delay.
	if(GetAndResWLANTick() >= WLAN_PSE) {					// last time we were here is longer ago than Power Save of WLAN
		first = 3;											// number of data bytes to be sent with delay
		Delay_us(270);										// wait after CS and before sending first byte	
	} else {
		first = 0;
	}
	if(InFlash) {
		while(pgm_read_byte(Com)) {							// until terminating 0 is found in string
			wait = 0;
			while(WLAN_SPIR_SET) {
				if(++wait >= 200) {							// wait max 200*1us = 200us for SPI_INT go down
					break;
				}
				Delay_us(1);								// delay one us
			}
			if(!pgm_read_byte(Com+1)) {						// next character is end of string
				WriteSPIInt(pgm_read_byte(Com));			// use flash string, SPI int is enabled in here!
				break;										// end transmission
			}
			WLANWriteByteSPI(pgm_read_byte(Com));			// use flash string
			Com++;											// next byte
			if(first) {										// wake up WLAN slooowly...
				first--;									// how many data bytes to delay
				Delay_us(270);								// delay between bytes
			}
		}
	} else {
		while(*Com) {
			wait = 0;
			while(WLAN_SPIR_SET) {							// wait for WLAN Mmodule ready
				if(++wait >= 200) {							// wait max 200*1us = 200us for SPI_INT go down
					break;
				}
				Delay_us(1);								// delay one us
			}
			if(!(*(Com+1))) {								// next character is end of string
				WriteSPIInt(*Com);							// use flash string, SPI int is enabled in here!
				break;										// end transmission
			}
			WLANWriteByteSPI(*Com);							// use data string
			Com++;											// next byte
			if(first) {										// wake up WLAN slooowly...
				first--;									// how many data bytes to delay
				Delay_us(270);								// delay between bytes
			}
		}
	}
	WLANCS_PORT |= WLAN_CS;									// deactivate WLAN SPI interface	
	RESET_WLAN_SPI;											// restore SPI Mode for other modules
	KM_RELEASE_SPI();										// release SPI interface but keep WLAN_BLOCK
}						

//********************************************************/
// WLANGetResp
// Reads the Response from a previous command.
// Expects the response to be terminated by 0x0D,0x0A
// 
// Parameters: 	Buffer to store response
//				Max length of response, rest is thrown away
// Returns:		0: nothing received, timeout, response too long
//				>0 ; length of data string, 0 terminated
//********************************************************/
int16_t WLANGetResp(char *Resp, uint16_t MaxLen)
{
	uint16_t	ii, len, index, lindex;
	uint8_t		err;

	index = lindex = 0;
	err = 0;
	
	while(1) {
		if(xSemaphoreTake(qSPIRSema, SPIR_TMEOUT_AT) != pdTRUE) {	// wait for response, short timeout for regular commands
			// timeout occured
			#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
			WLANTmeOut++;									// increase debug timeout counter
			#endif
			WLAN_DIS_SPIR_INT;								// disable int
			WLAN_RELEASE();									// release WLAN
			return -1;										// indicate error
		}
		WLAN_EN_SPIR_INT;									// enable SPI interrupt
 		KM_WAIT_SPI();										// wait for SPI bus available
		SET_WLAN_SPI;										// data available, get it!
		WLANCS_PORT &= ~WLAN_CS;							// activate WLAN SPI
		len = WLANWriteByteSPI(0);							// get first length byte
		if(!(len & 0x80)) {									// highest bit must be set, Readyv bit of WLAN Module
			WLANCS_PORT |= WLAN_CS;							// unexpected SPI int without data from WLAN
			RESET_WLAN_SPI;									// deactivate WLAN SPI, restore SPI Mode
			KM_RELEASE_SPI();								// release SPI bus
			WLAN_RELEASE();									// release WLAN
			#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
			WLANTmeOut++;									// increase debug timeout counter
			#endif
			return -1;										// report error
		}
		len = (((uint16_t)(len & 0x0F))<<8); 				// first byte of length
		len += WLANWriteByteSPI(0);							// here we have the lower portion of the length
		for(ii=0;ii<len;ii++) {								// get the data
			Resp[index] = WLANWriteByteSPI(0);				// get data byte
			if(Resp[index] == 0x0A) {						// parse for end of line
				if(!strncmp_P(&Resp[lindex], PSTR("I/OK"), 4)) {	// end of response found, quit
					Resp[index+1] = 0x00;					// terminate string
					WLANCS_PORT |= WLAN_CS;					// deactivate WLAN SPI
					RESET_WLAN_SPI;							// restore SPI Mode
					KM_RELEASE_SPI();
					WLAN_DIS_SPIR_INT;						// no interrupt required any more
					WLAN_RELEASE();							// ready for next command
					if(err) {
						#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
						WLANTmeOut++;									// increase debug timeout counter
						#endif
						return -1;
					} else {
						return index;						// return length of data
					}
				} 
				if(!strncmp_P(&Resp[lindex], PSTR("I/ERR"), 5)) {	// end of response found, quit
					Resp[index+1] = 0x00;					// terminate string
					WLANCS_PORT |= WLAN_CS;					// deactivate WLAN SPI
					RESET_WLAN_SPI;							// restore SPI Mode
					KM_RELEASE_SPI();
					WLAN_DIS_SPIR_INT;						// no interrupt required any more
					WLAN_RELEASE();							// ready for next command
					#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
					WLANTmeOut++;									// increase debug timeout counter
					#endif
					return -1;								// return length of data
				}
				if(index >= (MaxLen-20)) {					// enough buffer to store I/OK or I/ERROR (xxx) in next line?
					lindex = index = 0;						// no -> reset to begin of buffer
					err = 1;								// indicate error, this should not happen, buffer too small
				} else {
					lindex = ++index;						// save position of last received line
				}
			} else {										// no end of line
				if(index < (MaxLen - 2)) {					// end of buffer reached?
					index++;								// no, next byte ready to receive
				} else {									// end of buffer reached
					lindex = index = 0;						// reset all to beginn of buffer
					err = 1;								// indicate error as this should not happen
				}
			}
		}
		WLANCS_PORT |= WLAN_CS;								// deactivate WLAN SPI
		RESET_WLAN_SPI;										// restore SPI Mode
		KM_RELEASE_SPI();
	}
}

//********************************************************/
// WLANGetSLNK
// Reads the Response from a previous SLNK command.
// Responses of WiReach are quite different:
// 1. Everthing OK:
//	I/OK			- SLNK accepted (fast)
//  I/0				- response is there, could be 0 or a value indicating length
//  OK				- net data, whatever the requested web page respond (inkl. errors), can be loooong!!
//  I/ONLINE		- after a few seconds. Nice to find the termination
//
// 2. No connection
//	I/OK			- SLNK accepted (fast)
//	I/ERROR (301)	- no connection after about 35 seconds, no I/ONLINE afterwards!
// 
// Parameters: 	Buffer to store response
//				Max length of response, rest is thrown away
// Returns:		0: nothing received, timeout, response too long
//				>0 ; length of data string, 0 terminated
//********************************************************/
int16_t WLANGetSLNK(char *Resp, uint16_t MaxLen)
{
	uint16_t	ii, len, index, lindex;
	uint8_t		err;

	index = lindex = 0;
	err = 0;

	while(1) {
		if(xSemaphoreTake(qSPIRSema, SPIR_TMEOUT_CON) != pdTRUE) {	// wait for response, loong timeout >35s !!!
			// timeout occured
			#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
			WLANTmeOut++;									// increase debug timeout counter
			#endif
			WLAN_DIS_SPIR_INT;								// disable int
			WLAN_RELEASE();									// release WLAN
			return -1;										// indicate error
		}
		WLAN_EN_SPIR_INT;									// enable SPI interrupt
 		KM_WAIT_SPI();										// wait for SPI bus available
		SET_WLAN_SPI;										// data available, get it!
		WLANCS_PORT &= ~WLAN_CS;							// activate WLAN SPI
		len = WLANWriteByteSPI(0);							// get first length byte
		if(!(len & 0x80)) {									// highest bit must be set, Readyv bit of WLAN Module
			WLANCS_PORT |= WLAN_CS;							// unexpected SPI int without data from WLAN
			RESET_WLAN_SPI;									// deactivate WLAN SPI, restore SPI Mode
			KM_RELEASE_SPI();								// release SPI bus
			WLAN_RELEASE();									// release WLAN
			#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
			WLANTmeOut++;									// increase debug timeout counter
			#endif
			return -1;										// report error
		}
		len = (((uint16_t)(len & 0x0F))<<8); 				// first byte of length
		len += WLANWriteByteSPI(0);							// here we have the lower portion of the length
		for(ii=0;ii<len;ii++) {								// get the data
			Resp[index] = WLANWriteByteSPI(0);				// get data byte
			if(Resp[index] == 0x0A) {						// parse for end of line
				if(!strncmp_P(&Resp[lindex], PSTR("I/ONL"), 5)) {	// end of response found (I/ONLINE), quit
					Resp[index+1] = 0x00;					// terminate string
					WLANCS_PORT |= WLAN_CS;					// deactivate WLAN SPI
					RESET_WLAN_SPI;							// restore SPI Mode
					KM_RELEASE_SPI();
					WLAN_DIS_SPIR_INT;						// no interrupt required any more
					WLAN_RELEASE();							// ready for next command
					if(err) {
						#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
						WLANTmeOut++;									// increase debug timeout counter
						#endif
						return -1;
					} else {
						return index;						// return length of data
					}
				} 
				if(!strncmp_P(&Resp[lindex], PSTR("I/ERR"), 5)) {	// end of response found (I/ERROR (xxx)), quit
					Resp[index+1] = 0x00;					// terminate string
					WLANCS_PORT |= WLAN_CS;					// deactivate WLAN SPI
					RESET_WLAN_SPI;							// restore SPI Mode
					KM_RELEASE_SPI();
					WLAN_DIS_SPIR_INT;						// no interrupt required any more
					WLAN_RELEASE();							// ready for next command
					#if	(INCLUDE_uxTaskGetStackHighWaterMark == 1)
					WLANTmeOut++;									// increase debug timeout counter
					#endif
					return -1;								// return length of data
				}
				if(index >= (MaxLen-20)) {					// enough buffer to store I/OK or I/ERROR (xxx) in next line?
					lindex = index = 0;						// no -> reset to begin of buffer
					err = 1;								// indicate error, this should not happen, buffer too small
				} else {
					lindex = ++index;						// save position of last received line
				}
			} else {										// no end of line
				if(index < (MaxLen - 2)) {					// end of buffer reached?
					index++;								// no, next byte ready to receive
				} else {									// end of buffer reached
					lindex = index = 0;						// reset all to beginn of buffer
					err = 1;								// indicate error as this should not happen
				}
			}
		}
		WLANCS_PORT |= WLAN_CS;								// deactivate WLAN SPI
		RESET_WLAN_SPI;										// restore SPI Mode
		KM_RELEASE_SPI();
	}
}

//*************************************
// WLANWriteByteSPI
//
// Writes a single byte to SPI port
//
// Parameters:
// Data: Byte to write
//*************************************
uint8_t WLANWriteByteSPI(uint8_t Data)
{
	SPDR = Data;
	while(!(SPSR & (1<<SPIF))); 							// Wait byte off
	return SPDR;
}

//*************************************
// WriteSPIInt
//
// Write a single byte to SPI port and set SPI interrupt
//
// Parameters:
// Data: Byte to write
//*************************************
void WriteSPIInt(uint8_t Data)
{
	taskDISABLE_INTERRUPTS();								// send last byte under cli to ensure to get SPI int
	SPDR = Data;
	while(!(SPSR & (1<<SPIF))); 							// Wait byte off
	xSemaphoreTake(qSPIRSema,0);							// clear semaphore
	WLAN_EN_SPIR_INT;										// now enable SPI interrupt, late to avoid to get spikes from WLAN module
	taskENABLE_INTERRUPTS();								// allow interrupts again
}

//*************************************
// WLANSpiInit
//
// Initialises the SPI interface
//
// Parameters:
// none
//*************************************
void WLANSpiInit(void)
{
    SPCR = (0 << SPIE) | 									// SPI Interrupt Disable
           (1 << SPE)  | 									// SPI Enable
           (0 << DORD) | 									// Data Order: MSB first
           (1 << MSTR) | 									// Master mode
           (0 << CPOL) | 									// Clock Polarity: SCK low when idle
           (0 << CPHA) | 									// Clock Phase: sample on rising SCK edge
           (0 << SPR1) | 									// Clock Frequency: f_OSC / 4 = 2,7 MHz
           (0 << SPR0);
    SPSR |= (1 << SPI2X); 									// doubled clock frequency
															// now we have 5,4 MHz SPI Freq
}

